import math
import random
from typing import Any
from typing import Dict
from typing import List
from typing import Optional
from typing import Union

from ..logs.base import Logger
from ..market import Market
from ..order import LIMIT_ORDER
from ..order import Cancel
from ..order import Order
from ..utils.json_random import JsonRandom
from .base import Agent

MARGIN_FIXED = 0
MARGIN_NORMAL = 1


class FCNAgent(Agent):
    """FCN (Fundamental, Chartist, Noise) Agent class

    This class inherits from the :class:`pams.agents.Agent` class.

    An order decision mechanism proposed in Chiarella & Iori (2004).
    It employs two simple margin-based random tradings. Given an expected future price p, submit an order of price

    - "fixed" : :math:`p (1 ± k)` where :math:`0 \leq k \leq 1`
    - "normal" : :math:`p + N(0, k)` where :math:`k > 0`

    References:
        - Chiarella, C., & Iori, G. (2002). A simulation analysis of the microstructure of double auction markets.
          Quantitative Finance, 2(5), 346–353. https://doi.org/10.1088/1469-7688/2/5/303
    """  # NOQA

    fundamental_weight: float
    chart_weight: float
    margin_type: int
    mean_reversion_time: int
    noise_scale: float
    noise_weight: float
    order_margin: float
    time_window_size: int

    def __init__(
        self,
        agent_id: int,
        prng: random.Random,
        simulator: "Simulator",  # type: ignore  # NOQA
        name: str,
        logger: Optional[Logger] = None,
    ):
        super().__init__(agent_id, prng, simulator, name, logger)
        self.is_chart_following = True

    def is_finite(self, x: float) -> bool:
        """determine if it is a valid value.

        Args:
            x (float): value.

        Return:
            bool: whether or not it is a valid (not NaN, finite) value.
        """
        return not math.isnan(x) and not math.isinf(x)

    def setup(
        self,
        settings: Dict[str, Any],
        accessible_markets_ids: List[int],
        *args: Any,
        **kwargs: Any,
    ) -> None:
        """agent setup.  Usually be called from simulator/runner automatically.

        Args:
            settings (Dict[str, Any]): agent configuration. This can include the parameters "fundamentalWeight", "chartWeight",
                                       "noiseWeight", "noiseScale", "timeWindowSize", "orderMargin", "marginType",
                                       and "meanReversionTime".
            accessible_markets_ids (List[int]): list of market IDs.

        Returns:
            None
        """
        super().setup(settings=settings, accessible_markets_ids=accessible_markets_ids)
        json_random: JsonRandom = JsonRandom(prng=self.prng)
        self.fundamental_weight = json_random.random(
            json_value=settings["fundamentalWeight"]
        )
        self.chart_weight = json_random.random(json_value=settings["chartWeight"])
        self.noise_weight = json_random.random(json_value=settings["noiseWeight"])
        self.noise_scale = json_random.random(json_value=settings["noiseScale"])
        self.time_window_size = int(
            json_random.random(json_value=settings["timeWindowSize"])
        )
        self.order_margin = json_random.random(json_value=settings["orderMargin"])
        if settings.get("marginType") in [None, "fixed"]:
            self.margin_type = MARGIN_FIXED
        elif settings.get("marginType") == "normal":
            self.margin_type = MARGIN_NORMAL
        else:
            raise ValueError(
                "marginType have to be normal or fixed (not specified is also allowed.)"
            )
        if "meanReversionTime" in settings:
            self.mean_reversion_time = int(
                json_random.random(json_value=settings["meanReversionTime"])
            )
        else:
            self.mean_reversion_time = self.time_window_size

    def submit_orders(self, markets: List[Market]) -> List[Union[Order, Cancel]]:
        """submit orders based on FCN-based calculation.

        .. seealso::
            - :func:`pams.agents.Agent.submit_orders`
        """
        orders: List[Union[Order, Cancel]] = sum(
            [self.submit_orders_by_market(market=market) for market in markets], []
        )
        return orders

    def submit_orders_by_market(self, market: Market) -> List[Union[Order, Cancel]]:
        """submit orders by market (internal usage).

        Args:
            market (Market): market to order.

        Returns:
            List[Union[Order, Cancel]]: order list.
        """
        orders: List[Union[Order, Cancel]] = []
        if not self.is_market_accessible(market_id=market.market_id):
            return orders

        time: int = market.get_time()
        time_window_size: int = min(time, self.time_window_size)
        assert time_window_size >= 0
        assert self.fundamental_weight >= 0.0
        assert self.chart_weight >= 0.0
        assert self.noise_weight >= 0.0

        fundamental_scale: float = 1.0 / max(self.mean_reversion_time, 1)
        fundamental_log_return = fundamental_scale * math.log(
            market.get_fundamental_price() / market.get_market_price()
        )
        assert self.is_finite(fundamental_log_return)

        chart_scale: float = 1.0 / max(time_window_size, 1)
        chart_mean_log_return = chart_scale * math.log(
            market.get_market_price() / market.get_market_price(time - time_window_size)
        )
        assert self.is_finite(chart_mean_log_return)

        noise_log_return: float = self.noise_scale * self.prng.gauss(mu=0.0, sigma=1.0)
        assert self.is_finite(noise_log_return)

        expected_log_return: float = (
            1.0 / (self.fundamental_weight + self.chart_weight + self.noise_weight)
        ) * (
            self.fundamental_weight * fundamental_log_return
            + self.chart_weight
            * chart_mean_log_return
            * (1 if self.is_chart_following else -1)
            + self.noise_weight * noise_log_return
        )
        assert self.is_finite(expected_log_return)

        expected_future_price: float = market.get_market_price() * math.exp(
            expected_log_return * self.time_window_size
        )
        assert self.is_finite(expected_future_price)

        if self.margin_type == MARGIN_FIXED:
            assert 0.0 <= self.order_margin <= 1.0

            order_volume: int = 1

            if expected_future_price > market.get_market_price():
                order_price = expected_future_price * (1 - self.order_margin)
                orders.append(
                    Order(
                        agent_id=self.agent_id,
                        market_id=market.market_id,
                        is_buy=True,
                        kind=LIMIT_ORDER,
                        volume=order_volume,
                        price=order_price,
                        ttl=self.time_window_size,
                    )
                )
            if expected_future_price < market.get_market_price():
                order_price = expected_future_price * (1 + self.order_margin)
                orders.append(
                    Order(
                        agent_id=self.agent_id,
                        market_id=market.market_id,
                        is_buy=False,
                        kind=LIMIT_ORDER,
                        volume=order_volume,
                        price=order_price,
                        ttl=self.time_window_size,
                    )
                )

        if self.margin_type == MARGIN_NORMAL:
            assert self.order_margin >= 0.0
            order_price = (
                expected_future_price
                + self.prng.gauss(mu=0.0, sigma=1.0) * self.order_margin
            )
            order_volume = 1
            assert order_price >= 0.0
            assert order_volume > 0
            if expected_future_price > market.get_market_price():
                orders.append(
                    Order(
                        agent_id=self.agent_id,
                        market_id=market.market_id,
                        is_buy=True,
                        kind=LIMIT_ORDER,
                        volume=order_volume,
                        price=order_price,
                        ttl=self.time_window_size,
                    )
                )
            if expected_future_price < market.get_market_price():
                orders.append(
                    Order(
                        agent_id=self.agent_id,
                        market_id=market.market_id,
                        is_buy=False,
                        kind=LIMIT_ORDER,
                        volume=order_volume,
                        price=order_price,
                        ttl=self.time_window_size,
                    )
                )
        return orders

    def __repr__(self) -> str:
        """string representation of FCN agent class.

        Returns:
            str: string representation of this class.
        """
        return (
            f"<{self.__class__.__module__}.{self.__class__.__name__} | id={self.agent_id}, rnd={self.prng}, "
            f"chart_weight={self.chart_weight}, fundamental_weight={self.fundamental_weight}, "
            f"noise_weight={self.noise_weight}, is_chart_following:{self.is_chart_following}, "
            f"margin_type={self.margin_type}, mean_reversion_time:{self.mean_reversion_time}, "
            f"noise_scale={self.noise_scale}, time_window_size={self.time_window_size}, "
            f"order_margin={'MARGIN_FIXED' if self.order_margin == MARGIN_FIXED else 'MARGIN_NORMAL'}"
        )



import json
import math
import random
from typing import Dict
from typing import List
from typing import Union

JsonValue = Union[Dict, List, float, int]


class JsonRandom:
    """random generator from json.

    The following direction can be used for config as randomized values:
     - :code:`[a, b]`: uniform distribution started from a and ended to b. Not that the value should be int, this automatically converted into int. This always satisfy :math:`a \leq x < b`
     - :code:`{"const": [a]}`: constant value. Always set to a.
     - :code:`{"uniform": [a, b]}`: same as [a, b]
     - :code:`{"normal": [u, s]}`: normal distribution whose mean and deviation is u and s.
     - :code:`{"expon": [lam]}`: exponential distribution whose mean and deviation is lam.

    Examples:
        >>> from pams.utils.json_random import JsonRandom
        >>> import random
        >>> jr = JsonRandom(prng=random.Random(42))
        >>> [jr.random([10, 20]) for x in range(10)]
        [16.39426798457884, 10.25010755222667, 12.750293183691193, 12.232107381488227, 17.364712141640126, 16.766994874229113, 18.921795677048454, 10.869388326294162, 14.219218196852704, 10.297972194380703]
        >>> [jr.random({"const": [10]}) for x in range(10)]
        [10.0, 10.0, 10.0, 10.0, 10.0, 10.0, 10.0, 10.0, 10.0, 10.0]
        >>> [jr.random({"uniform": [10, 20]}) for x in range(10)]
        [12.186379748036034, 15.053552881033625, 10.265359696838637, 11.988376506866485, 16.49884437779523, 15.449414806032166, 12.204406220406966, 15.892656838759088, 18.094304566778266, 10.064987596780611]
        >>> [jr.random({"normal": [0, 1]}) for x in range(10)]
        [0.5317762204008692, -1.453545298008678, -0.3122773171445598, 0.49036253259352475, 0.8734043853794468, -0.2406296726551354, 0.3765998586879102, 0.24821344932841446, 0.7823268087036421, -1.1132222142481727]
        >>> [jr.random({"expon": [3]}) for x in range(10)]
        [0.642818017709456, 0.9452346835236866, 1.869586994011895, 0.08175668259806873, 2.9143451561160503, 1.7824008841046926, 0.5611413226153803, 1.4412784552296345, 0.4465202669419299, 1.6479086846872075]
    """  # NOQA

    def __init__(self, prng: random.Random) -> None:
        """initialization.

        Args:
            prng (random.Random): pseudo random number generator for this event.

        Returns:
            None
        """
        self.prng: random.Random = prng

    def _next_uniform(self, min_value: float, max_value: float) -> float:
        """get next uniform.

        Its probability density function is
        :math:`p(x) = \frac{1}{max - min}`
        anywhere within the interval :math:`[min, max)`, and 0 elsewhere.

        Args:
            min_value (float): min value.
            max_value (float): max value.

        Returns:
            float: uniform.
        """
        return self.prng.random() * (max_value - min_value) + min_value

    def _next_normal(self, mu: float, sigma: float) -> float:
        """get next normal.

        Its probability density function is
        :math:`p(x) = \frac{1}{\\sqrt{2 \\pi \\sigma^2}e^{- \frac{(x - \\mu)^2}{2 \\sigma^2}}}`
        where :math:`\\mu` is the mean and :math:`\\sigma` is the standard deviation.

        Args:
            mu (float): mu.
            sigma (float): sigma.

        Returns:
            float: normal.
        """
        return self.prng.gauss(mu=mu, sigma=sigma)

    def _next_exponential(self, lam: float) -> float:
        """get next exponential.

        Its probability density function is
        :math:`p(x) = \\lambda exp(- \\lambda x)`
        for :math:`x \\gt 0` and 0 elsewhere. Where :math:`\\lambda` is the scale parameter.

        Args:
            lam (float): lambda.

        Returns:
            float: exponential.
        """
        return lam * -math.log(self.prng.random())

    def random(self, json_value: JsonValue) -> float:
        """get a random value.

        Args:
            json_value (JsonValue): random type. This can include the parameter "const", "uniform", "normal", and "expon".

        Returns:
            float: random value.
        """
        if isinstance(json_value, list):
            if len(json_value) != 2:
                raise ValueError(
                    "Uniform distribution must be [min, max] but "
                    + json.dumps(json_value)
                )
            min_value: float = float(json_value[0])
            max_value: float = float(json_value[1])
            return self._next_uniform(min_value=min_value, max_value=max_value)
        if isinstance(json_value, dict):
            if len(json_value) != 1:
                raise ValueError(
                    "Multiple specification of distribution type: "
                    + json.dumps(json_value)
                )
            if "const" in json_value:
                args = json_value["const"]
                if not isinstance(args, list):
                    raise ValueError(
                        "Constant must be [value] (list) but " + json.dumps(json_value)
                    )
                if len(args) != 1:
                    raise ValueError(
                        "Constant must be [value] but " + json.dumps(json_value)
                    )
                value = float(args[0])
                return value
            if "uniform" in json_value:
                args = json_value["uniform"]
                if not isinstance(args, list):
                    raise ValueError(
                        "Uniform distribution must be [min, max] (list) but "
                        + json.dumps(json_value)
                    )
                if len(args) != 2:
                    raise ValueError(
                        "Uniform distribution must be [min, max] but "
                        + json.dumps(json_value)
                    )
                min_value = float(args[0])
                max_value = float(args[1])
                return self._next_uniform(min_value=min_value, max_value=max_value)
            if "normal" in json_value:
                args = json_value["normal"]
                if not isinstance(args, list):
                    raise ValueError(
                        "Normal distribution must be [mu, sigma] (list) but "
                        + json.dumps(json_value)
                    )
                if len(args) != 2:
                    raise ValueError(
                        "Normal distribution must be [mu, sigma] but "
                        + json.dumps(json_value)
                    )
                mu = float(args[0])
                sigma = float(args[1])
                return self._next_normal(mu=mu, sigma=sigma)
            if "expon" in json_value:
                args = json_value["expon"]
                if not isinstance(args, list):
                    raise ValueError(
                        "Exponential distribution must be [lambda] (list) but "
                        + json.dumps(json_value)
                    )
                if len(args) != 1:
                    raise ValueError(
                        "Exponential distribution must be [lambda] but "
                        + json.dumps(json_value)
                    )
                lam = float(args[0])
                return self._next_exponential(lam=lam)
            raise ValueError("Unknown distribution type: " + json.dumps(json_value))
        return float(json_value)



from .base import EventABC
from .base import EventHook
from .fundamental_price_shock import FundamentalPriceShock
from .price_limit_rule import PriceLimitRule
from .trading_halt_rule import TradingHaltRule



import random
import warnings
from typing import Any
from typing import Dict
from typing import Optional

from .logs.base import Logger


class Session:
    """Session management class."""

    def __init__(
        self,
        session_id: int,
        prng: random.Random,
        session_start_time: int,
        simulator: "Simulator",  # type: ignore  # NOQA
        name: str,
        logger: Optional[Logger] = None,
    ) -> None:
        """initialization.

        Args:
            session_id (int): session ID.
            prng (random.Random): pseudo random number generator for this session.
            session_start_time (int): start time of this session.
            simulator (Simulator): this is used for accessing simulation environment.
            name (str): session name.
            logger (Logger, Optional): logger for correcting various outputs in one simulation.
                                       logger is usually shared to all classes.
                                       Please note that logger is usually not thread-safe and non-blocking.

        Returns:
            None

        Note:
             `prng` should not be shared with other classes and be used only in this class.
             It is because sometimes agent process runs one of parallelized threads.
        """
        self.session_id: int = session_id
        self.name: str = name
        self.prng: random.Random = prng
        self.simulator: "Simulator" = simulator  # type: ignore  # NOQA
        self.logger: Optional[Logger] = logger

        self.iteration_steps: int = 0
        self.max_high_frequency_orders: int = 1
        self.max_normal_orders: int = 1
        self.with_order_execution: bool = False
        self.with_order_placement: bool = False
        self.with_print: bool = True
        self.high_frequency_submission_rate: float = 1.0
        self.session_start_time: int = session_start_time

    def __repr__(self) -> str:
        return (
            f"<{self.__class__.__module__}.{self.__class__.__name__} | id={self.session_id}, name={self.name}, "
            f"iteration_steps={self.iteration_steps}, session_start_time={self.session_start_time}, "
            f"max_normal_orders={self.max_normal_orders}, max_high_frequency_orders={self.max_high_frequency_orders}, "
            f"with_order_placement={self.with_order_placement}, with_order_execution={self.with_order_execution}, "
            f"high_frequency_submission_rate={self.high_frequency_submission_rate}, with_print={self.with_print}, "
            f"logger={self.logger.__str__()}>"
        )

    def setup(self, settings: Dict[str, Any], *args, **kwargs) -> None:  # type: ignore  # NOQA
        """setup session configuration from setting format.

        Args:
            settings (Dict[str, Any]): session configuration. Usually, automatically set from json config of simulator.
                                       This must include the parameters "iterationSteps", "withOrderPlacement", "withOrderExecution", and "withPrint".
                                       This can include the parameter "maxNormalOrders", "maxHighFrequencyOrders", and "highFrequencySubmitRate".

        Returns:
            None
        """
        if "iterationSteps" not in settings:
            raise ValueError(
                "for each element in simulation.sessions must have iterationSteps"
            )
        if not isinstance(settings["iterationSteps"], int):
            raise ValueError("iterationSteps must be int")
        self.iteration_steps = int(settings["iterationSteps"])
        if "withOrderPlacement" not in settings:
            raise ValueError(
                "for each element in simulation.sessions must have withOrderPlacement"
            )
        if not isinstance(settings["withOrderPlacement"], bool):
            raise ValueError("withOrderPlacement must be boolean")
        self.with_order_placement = settings["withOrderPlacement"]
        if "withOrderExecution" not in settings:
            raise ValueError(
                "for each element in simulation.sessions must have withOrderExecution"
            )
        if not isinstance(settings["withOrderExecution"], bool):
            raise ValueError("withOrderExecution must be boolean")
        self.with_order_execution = settings["withOrderExecution"]
        if "withPrint" not in settings:
            raise ValueError(
                "for each element in simulation.sessions must have withPrint"
            )
        if not isinstance(settings["withPrint"], bool):
            raise ValueError("withPrint must be boolean")
        self.with_print = settings["withPrint"]
        if "maxNormalOrders" in settings:
            # TODO: check non-negative
            self.max_normal_orders = settings["maxNormalOrders"]
        if "maxHighFrequencyOrders" in settings:
            # TODO: check non-negative
            self.max_high_frequency_orders = settings["maxHighFrequencyOrders"]
            if "maxHifreqOrders" in settings:
                raise ValueError(
                    "maxHifreqOrders is replaced to maxHighFrequencyOrders in pams. Please delete it."
                )
        elif "maxHifreqOrders" in settings:
            # TODO: check non-negative
            warnings.warn(
                "maxHifreqOrders is replaced to maxHighFrequencyOrders in pams."
            )
            self.max_high_frequency_orders = settings["maxHifreqOrders"]
        # TODO: check malOrders + maxHighFrequencyOrders >= 1
        if "highFrequencySubmitRate" in settings:
            # TODO: check non-negative
            self.high_frequency_submission_rate = settings["highFrequencySubmitRate"]
            if "hifreqSubmitRate" in settings:
                raise ValueError(
                    "hifreqSubmitRate is replaced to highFrequencySubmitRate in pams. Please delete it."
                )
        elif "hifreqSubmitRate" in settings:
            warnings.warn(
                "hifreqSubmitRate is replaced to highFrequencySubmitRate in pams."
            )
            # TODO: check non-negative
            self.max_high_frequency_orders = settings["hifreqSubmitRate"]



import random
from typing import Any
from typing import Dict
from typing import List
from typing import Optional
from typing import Union

from ..index_market import IndexMarket
from ..logs import Logger
from ..market import Market
from ..order import LIMIT_ORDER
from ..order import Cancel
from ..order import Order
from .high_frequency_agent import HighFrequencyAgent


class ArbitrageAgent(HighFrequencyAgent):
    """Arbitrage Agent class.

    This class inherits from the HighFrequencyAgent class.
    Arbitrage agent mainly aimed to take arbitrage chance between spot markets and its index market.

    Note:
        Currently, index markets must have the same weight for each constitutional stock.
    """

    def __init__(
        self,
        agent_id: int,
        prng: random.Random,
        simulator: "Simulator",  # type: ignore  # NOQA
        name: str,
        logger: Optional[Logger] = None,
    ) -> None:
        super().__init__(
            agent_id=agent_id, prng=prng, simulator=simulator, name=name, logger=logger
        )

        self.order_volume: int = 1
        self.order_threshold_price: float = 1.0
        self.order_time_length: int = 1

    def setup(  # type: ignore
        self,
        settings: Dict[str, Any],
        accessible_markets_ids: List[int],
        *args,
        **kwargs,
    ) -> None:
        """agent setup. Usually be called from simulator/runner automatically.

        Args:
            settings (Dict[str, Any]): agent configuration. Usually, automatically set from json config of simulator.
                                       This must include the parameters "orderVolume", "orderThresholdPrice".
                                       This can include the parameter "orderTimeLength".
            accessible_markets_ids (List[int]): list of market IDs.

        Returns:
            None
        """
        super(ArbitrageAgent, self).setup(
            settings, accessible_markets_ids, *args, **kwargs
        )
        if "orderVolume" not in settings:
            raise ValueError("orderVolume is required for ArbitrageAgent")
        if not isinstance(settings["orderVolume"], int):
            raise ValueError("orderVolume have to be int")
        self.order_volume = settings["orderVolume"]
        if "orderThresholdPrice" not in settings:
            raise ValueError("orderThresholdPrice is required for ArbitrageAgent")
        self.order_threshold_price = settings["orderThresholdPrice"]
        if "orderTimeLength" in settings:
            if not isinstance(settings["orderTimeLength"], int):
                raise ValueError("orderTimeLength have to be int")
            self.order_time_length = settings["orderTimeLength"]

    def _submit_orders(self, market: Market) -> List[Union[Order, Cancel]]:
        """internal sub routine for submitting orders by market.

        Args:
            market (List[Market]): markets to order.

        Returns:
            List[Union[Order, Cancel]]: order list.
        """
        orders: List[Union[Order, Cancel]] = []
        if not isinstance(market, IndexMarket):
            return orders
        if not self.is_market_accessible(market_id=market.market_id):
            return orders
        index: IndexMarket = market
        spots: List[Market] = index.get_components()
        if not index.is_running or not index.is_all_markets_running():
            return orders
        market_index: float = index.get_index()
        market_price: float = index.get_market_price()

        if len(set(map(lambda x: x.outstanding_shares, spots))) > 1:
            raise NotImplementedError(
                "currently, the components must have the same outstanding shares"
            )

        if (
            market_price < market_index
            and market_index - market_price > self.order_threshold_price
        ):
            index_order_volume = len(spots) * self.order_volume
            orders.append(
                Order(
                    agent_id=self.agent_id,
                    market_id=index.market_id,
                    is_buy=True,
                    kind=LIMIT_ORDER,
                    volume=index_order_volume,
                    price=index.get_market_price(),
                    ttl=self.order_time_length,
                )
            )
            for m in spots:
                orders.append(
                    Order(
                        agent_id=self.agent_id,
                        market_id=m.market_id,
                        is_buy=False,
                        kind=LIMIT_ORDER,
                        volume=self.order_volume,
                        price=m.get_market_price(),
                        ttl=self.order_time_length,
                    )
                )
        if (
            market_price > market_index
            and market_price - market_index > self.order_threshold_price
        ):
            index_order_volume = len(spots) * self.order_volume
            orders.append(
                Order(
                    agent_id=self.agent_id,
                    market_id=index.market_id,
                    is_buy=False,
                    kind=LIMIT_ORDER,
                    volume=index_order_volume,
                    price=index.get_market_price(),
                    ttl=self.order_time_length,
                )
            )
            for m in spots:
                orders.append(
                    Order(
                        agent_id=self.agent_id,
                        market_id=m.market_id,
                        is_buy=True,
                        kind=LIMIT_ORDER,
                        volume=self.order_volume,
                        price=m.get_market_price(),
                        ttl=self.order_time_length,
                    )
                )
        return orders

    def submit_orders(self, markets: List[Market]) -> List[Union[Order, Cancel]]:
        """submit orders to take arbitrage chance.

        .. seealso::
            - :func:`pams.agents.Agent.submit_orders`
        """
        orders: List[Union[Order, Cancel]] = []
        for market in markets:
            orders.extend(self._submit_orders(market=market))
        return orders

    def __repr__(self) -> str:
        """string representation of FCN agent class.

        Returns:
            str: string representation of this class.
        """
        return (
            f"<{self.__class__.__module__}.{self.__class__.__name__} | id={self.agent_id}, rnd={self.prng}, "
            f"order_volume={self.order_volume}, order_threshold_price={self.order_threshold_price}, "
            f"order_time_length={self.order_time_length}>"
        )



import heapq
import math
import random
import warnings
from typing import Any
from typing import Dict
from typing import Iterable
from typing import List
from typing import Optional
from typing import Tuple
from typing import TypeVar
from typing import Union
from typing import cast

from .logs.base import CancelLog
from .logs.base import ExecutionLog
from .logs.base import Log
from .logs.base import Logger
from .logs.base import OrderLog
from .order import Cancel
from .order import Order
from .order_book import OrderBook

T = TypeVar("T")


class Market:
    """Market class.

    .. seealso::
        - :class:`pams.index_market.IndexMarket`: IndexMarket
    """

    def __init__(
        self,
        market_id: int,
        prng: random.Random,
        simulator: "Simulator",  # type: ignore  # NOQA
        name: str,
        logger: Optional[Logger] = None,
    ) -> None:
        """initialization.

        Args:
            market_id (int): market ID.
            prng (random.Random): pseudo random number generator for this market.
            simulator (:class:`pams.Simulator`): simulator that executes this market.
            name (str): market name.
            logger (Logger, Optional): logger.

        Returns:
            None
        """
        self.market_id: int = market_id
        self._prng = prng
        self.logger: Optional[Logger] = logger
        self._is_running: bool = False
        self.tick_size: float = 1.0
        self.chunk_size = 100
        self.sell_order_book: OrderBook = OrderBook(is_buy=False)
        self.buy_order_book: OrderBook = OrderBook(is_buy=True)
        self.time: int = -1
        self._market_prices: List[Optional[float]] = []
        self._last_executed_prices: List[Optional[float]] = []
        self._mid_prices: List[Optional[float]] = []
        self._fundamental_prices: List[Optional[float]] = []
        self._executed_volumes: List[int] = []
        self._executed_total_prices: List[float] = []
        self._n_buy_orders: List[int] = []
        self._n_sell_orders: List[int] = []
        self._next_order_id: int = 0
        self.simulator: "Simulator" = simulator  # type: ignore  # NOQA
        self.name: str = name
        self.outstanding_shares: Optional[int] = None

    def __repr__(self) -> str:
        return (
            f"<{self.__class__.__module__}.{self.__class__.__name__} | id={self.market_id}, name={self.name}, "
            f"tick_size={self.tick_size}, outstanding_shares={self.outstanding_shares}>"
        )

    def setup(self, settings: Dict[str, Any], *args, **kwargs) -> None:  # type: ignore  # NOQA
        """setup market configuration from setting format.

        Args:
            settings (Dict[str, Any]): market configuration. Usually, automatically set from json config of simulator.
                                       This must include the parameters "tickSize" and either "marketPrice" or "fundamentalPrice".
                                       This can include the parameter "outstandingShares".

        Returns:
            None
        """
        if "tickSize" not in settings:
            raise ValueError("tickSize is required")
        self.tick_size = settings["tickSize"]
        if "outstandingShares" in settings:
            if not isinstance(settings["outstandingShares"], int):
                raise ValueError("outstandingShares must be int")
            self.outstanding_shares = settings["outstandingShares"]
        if "marketPrice" in settings:
            self._market_prices = [float(settings["marketPrice"])]
        elif "fundamentalPrice" in settings:
            self._market_prices = [float(settings["fundamentalPrice"])]
        else:
            raise ValueError("fundamentalPrice or marketPrice is required for market")

    def _extract_sequential_data_by_time(
        self,
        times: Union[Iterable[int], None],
        parameters: List[Optional[T]],
        allow_none: bool = False,
    ) -> List[Optional[T]]:
        """extract sequential parameters by time. (Internal method)

        Args:
            times (Union[Iterable[int], None]): range of time steps.
            parameters (List[Optional[T]]): referenced parameters.
            allow_none (bool): whether a None result can be returned.

        Returns:
            List[Optional[T]]: extracted parameters.
        """
        if times is None:
            times = range(self.time + 1)
        if sum([t > self.time for t in times]) > 0:
            raise AssertionError("Cannot refer the future parameters")
        result = [parameters[t] for t in times]
        if not allow_none and None in result:
            raise AssertionError
        return result

    def _extract_data_by_time(
        self,
        time: Union[int, None],
        parameters: List[Optional[T]],
        allow_none: bool = False,
    ) -> Optional[T]:
        """extract a parameter by time. (Internal method)

        Args:
            time (Union[int, None]): time step.
            parameters (List[Optional[T]]): referenced parameters.
            allow_none (bool): whether a None result can be returned.

        Returns:
            Optional[T]: extracted parameter.
        """
        if time is None:
            time = self.time
        if time > self.time:
            raise AssertionError("Cannot refer the future parameters")
        result = parameters[time]
        if not allow_none and result is None:
            raise AssertionError
        return result

    def get_time(self) -> int:
        """get time step."""
        return self.time

    def get_market_prices(
        self, times: Union[Iterable[int], None] = None
    ) -> List[float]:
        """get market prices.

        Args:
            times (Union[Iterable[int], None]): range of time steps.

        Returns:
            List[float]: extracted sequential data.
        """
        return cast(
            List[float],
            self._extract_sequential_data_by_time(
                times, self._market_prices, allow_none=False
            ),
        )

    def get_market_price(self, time: Union[int, None] = None) -> float:
        """get market price.

        Args:
            time (Union[int, None]): time step.

        Returns:
            float: extracted data.
        """
        return cast(
            float,
            self._extract_data_by_time(time, self._market_prices, allow_none=False),
        )

    def get_mid_prices(
        self, times: Union[Iterable[int], None] = None
    ) -> List[Optional[float]]:
        """get middle prices.

        Args:
            times (Union[Iterable[int], None]): time steps.

        Returns:
            List[Optional[float]]: middle prices.
        """
        return self._extract_sequential_data_by_time(
            times, self._mid_prices, allow_none=True
        )

    def get_mid_price(self, time: Union[int, None] = None) -> Optional[float]:
        """get middle price.

        Args:
            time (Union[int, None]): time step.

        Returns:
            float, Optional: middle price.
        """
        return self._extract_data_by_time(time, self._mid_prices, allow_none=True)

    def get_last_executed_prices(
        self, times: Union[Iterable[int], None] = None
    ) -> List[Optional[float]]:
        """get prices executed last steps.

        Args:
            times (Union[Iterable[int], None]): time steps.

        Returns:
            List[Optional[float]]: prices.
        """
        return self._extract_sequential_data_by_time(
            times, self._last_executed_prices, allow_none=True
        )

    def get_last_executed_price(self, time: Union[int, None] = None) -> Optional[float]:
        """get price executed last step.

        Args:
            time (Union[int, None]): time step.

        Returns:
            float, Optional: price.
        """
        return self._extract_data_by_time(
            time, self._last_executed_prices, allow_none=True
        )

    def get_fundamental_prices(
        self, times: Union[Iterable[int], None] = None
    ) -> List[float]:
        """get fundamental prices.

        Args:
            times (Union[Iterable[int], None]): time steps.

        Returns:
            List[float]: fundamental prices.
        """
        return cast(
            List[float],
            self._extract_sequential_data_by_time(times, self._fundamental_prices),
        )

    def get_fundamental_price(self, time: Union[int, None] = None) -> float:
        """get fundamental price.

        Args:
            time (Union[int, None]): time step.

        Returns:
            float: fundamental price.
        """
        return cast(float, self._extract_data_by_time(time, self._fundamental_prices))

    def get_executed_volumes(
        self, times: Union[Iterable[int], None] = None
    ) -> List[int]:
        """get executed volumes.

        Args:
            times (Union[Iterable[int], None]): time steps.

        Returns:
            List[int]: volumes.
        """
        return cast(
            List[int],
            self._extract_sequential_data_by_time(
                times, cast(List[Optional[int]], self._executed_volumes)
            ),
        )

    def get_executed_volume(self, time: Union[int, None] = None) -> int:
        """get executed volume.

        Args:
            time (Union[int, None]): time step.

        Returns:
            int: volume.
        """
        return cast(
            int,
            self._extract_data_by_time(
                time, cast(List[Optional[int]], self._executed_volumes)
            ),
        )

    def get_executed_total_prices(
        self, times: Union[Iterable[int], None] = None
    ) -> List[float]:
        """get executed total prices.

        Args:
            times (Union[Iterable[int], None]): time steps.

        Returns:
            List[float]: total prices.
        """
        return cast(
            List[float],
            self._extract_sequential_data_by_time(
                times, cast(List[Optional[int]], self._executed_total_prices)
            ),
        )

    def get_executed_total_price(self, time: Union[int, None] = None) -> float:
        """get executed total price.

        Args:
            time (Union[int, None]): time step.

        Returns:
            float: total price.
        """
        return cast(
            float,
            self._extract_data_by_time(
                time, cast(List[Optional[int]], self._executed_total_prices)
            ),
        )

    def get_n_buy_orders(self, times: Union[Iterable[int], None] = None) -> List[int]:
        """get the number of buy orders.

        Args:
            times (Union[Iterable[int], None]): time steps.

        Returns:
            List[int]: number of buy orders.
        """
        return cast(
            List[int],
            self._extract_sequential_data_by_time(
                times, cast(List[Optional[int]], self._n_buy_orders)
            ),
        )

    def get_n_buy_order(self, time: Union[int, None] = None) -> int:
        """get the number of buy order.

        Args:
            time (Union[int, None]): time step.

        Returns:
            int: number of buy order.
        """
        return cast(
            int,
            self._extract_data_by_time(
                time, cast(List[Optional[int]], self._n_buy_orders)
            ),
        )

    def get_n_sell_orders(self, times: Union[Iterable[int], None] = None) -> List[int]:
        """get the number of sell orders.

        Args:
            times (Union[Iterable[int], None]): time steps.

        Returns:
            List[int]: number of sell orders.
        """
        return cast(
            List[int],
            self._extract_sequential_data_by_time(
                times, cast(List[Optional[int]], self._n_sell_orders)
            ),
        )

    def get_n_sell_order(self, time: Union[int, None] = None) -> int:
        """get the number of sell order.

        Args:
            time (Union[int, None]): time step.

        Returns:
            int: number of sell order.
        """
        return cast(
            int,
            self._extract_data_by_time(
                time, cast(List[Optional[int]], self._n_sell_orders)
            ),
        )

    def _fill_until(self, time: int) -> None:
        if len(self._mid_prices) >= time + 1:
            return
        length = (time // self.chunk_size + 1) * self.chunk_size
        self._market_prices = self._market_prices + [
            None for _ in range(length - len(self._market_prices))
        ]
        self._mid_prices = self._mid_prices + [
            None for _ in range(length - len(self._mid_prices))
        ]
        self._last_executed_prices = self._last_executed_prices + [
            None for _ in range(length - len(self._last_executed_prices))
        ]
        self._fundamental_prices = self._fundamental_prices + [
            None for _ in range(length - len(self._fundamental_prices))
        ]
        self._executed_volumes = self._executed_volumes + [
            0 for _ in range(length - len(self._executed_volumes))
        ]
        self._executed_total_prices = self._executed_total_prices + [
            0 for _ in range(length - len(self._executed_total_prices))
        ]
        self._n_buy_orders = self._n_buy_orders + [
            0 for _ in range(length - len(self._n_buy_orders))
        ]
        self._n_sell_orders = self._n_sell_orders + [
            0 for _ in range(length - len(self._n_sell_orders))
        ]

    def get_vwap(self, time: Optional[int] = None) -> float:
        """get VWAP.

        Args:
            time (int, Optional): time step.

        Returns:
            float: VWAP.
        """
        if time is None:
            time = self.time
        if time > self.time:
            raise AssertionError("Cannot refer the future parameters")
        if sum(self._executed_volumes[: time + 1]) == 0:
            return float("nan")
        return sum(self._executed_total_prices[: time + 1]) / sum(
            self._executed_volumes[: time + 1]
        )

    @property
    def is_running(self) -> bool:
        """get whether this market is running or not.

        Returns:
            bool: whether this market is running or not.
        """
        return self._is_running

    def get_best_buy_price(self) -> Optional[float]:
        """get the best buy price.

        Returns:
            float, Optional: the best buy price.
        """
        return self.buy_order_book.get_best_price()

    def get_best_sell_price(self) -> Optional[float]:
        """get the best sell price.

        Returns:
            float, Optional: the best sell price.
        """
        return self.sell_order_book.get_best_price()

    def get_sell_order_book(self) -> Dict[Optional[float], int]:
        """get sell order book.

        Returns:
            Dict[Optional[float], int]: sell order book.
        """
        return self.sell_order_book.get_price_volume()

    def get_buy_order_book(self) -> Dict[Optional[float], int]:
        """get buy order book.

        Returns:
            Dict[Optional[float], int]: buy order book.
        """
        return self.buy_order_book.get_price_volume()

    def convert_to_tick_level_rounded_lower(self, price: float) -> int:
        """convert price to tick level rounded lower.

        Args:
            price (float): price.

        Returns:
            int: price for tick level rounded lower.
        """
        return math.floor(price / self.tick_size)

    def convert_to_tick_level_rounded_upper(self, price: float) -> int:
        """convert price to tick level rounded upper.

        Args:
            price (float): price.

        Returns:
            int: price for tick level rounded upper.
        """
        return math.ceil(price / self.tick_size)

    def convert_to_tick_level(self, price: float, is_buy: bool) -> int:
        """convert price to tick level. If it is buy order price, it is rounded lower. If it is sell order price, it is rounded upper.

        Args:
            price (float): price.
            is_buy (bool): buy order or not.

        Returns:
            int: price for tick level.
        """
        if is_buy:
            return self.convert_to_tick_level_rounded_lower(price=price)
        else:
            return self.convert_to_tick_level_rounded_upper(price=price)

    def convert_to_price(self, tick_level: int) -> float:
        """convert tick to price.

        Args:
            tick_level (int): tick level.

        Returns:
            float: price.
        """
        return self.tick_size * tick_level

    def _set_time(self, time: int, next_fundamental_price: float) -> None:
        """set time step. (Usually, only triggered by simulator)

        Args:
            time (int): time step.
            next_fundamental_price (float): next fundamental price.

        Returns:
            None
        """
        self.time = time
        self.buy_order_book._set_time(time)
        self.sell_order_book._set_time(time)
        self._fill_until(time=time)
        self._fundamental_prices[self.time] = next_fundamental_price
        if self.time > 0:
            executed_prices: List[float] = cast(
                List[float],
                list(
                    filter(
                        lambda x: x is not None, self._last_executed_prices[: self.time]
                    )
                ),
            )
            self._last_executed_prices[self.time] = (
                executed_prices[-1] if sum(executed_prices) > 0 else None
            )
            mid_prices: List[float] = cast(
                List[float],
                list(filter(lambda x: x is not None, self._mid_prices[: self.time])),
            )
            self._mid_prices[self.time] = (
                mid_prices[-1] if sum(mid_prices) > 0 else None
            )
            market_prices: List[float] = cast(
                List[float],
                list(filter(lambda x: x is not None, self._market_prices[: self.time])),
            )
            self._market_prices[self.time] = (
                market_prices[-1] if sum(market_prices) > 0 else None
            )
            if self.is_running:
                if self._last_executed_prices[self.time - 1] is not None:
                    self._market_prices[self.time] = self._last_executed_prices[
                        self.time
                    ]
                elif self._mid_prices[self.time - 1] is not None:
                    self._market_prices[self.time] = self._mid_prices[self.time]

    def _update_time(self, next_fundamental_price: float) -> None:
        """update time. (Usually, only triggered by simulator)

        Args:
            next_fundamental_price (float): next fundamental price.

        Returns:
            None
        """
        self.time += 1
        self.buy_order_book._set_time(self.time)
        self.sell_order_book._set_time(self.time)
        self._fill_until(time=self.time)
        self._fundamental_prices[self.time] = next_fundamental_price
        if self.time > 0:
            self._last_executed_prices[self.time] = self._last_executed_prices[
                self.time - 1
            ]
            self._mid_prices[self.time] = self._mid_prices[self.time - 1]
            self._market_prices[self.time] = self._market_prices[self.time - 1]
            if self.is_running:
                if self._last_executed_prices[self.time - 1] is not None:
                    self._market_prices[self.time] = self._last_executed_prices[
                        self.time - 1
                    ]
                elif self._mid_prices[self.time - 1] is not None:
                    self._market_prices[self.time] = self._mid_prices[self.time - 1]
        else:
            if self._market_prices[self.time] is None:
                self._market_prices[self.time] = next_fundamental_price

    def _cancel_order(self, cancel: Cancel) -> CancelLog:
        """cancel order. (Usually, only triggered by simulator)

        Args:
            cancel (:class:`pams.order.Cancel`): cancel class.

        Returns:
            :class:`pams.logs.base.CancelLog`: cancel log.
        """
        if self.market_id != cancel.order.market_id:
            raise ValueError("this cancel order is for a different market")
        if cancel.order.order_id is None or cancel.order.placed_at is None:
            raise ValueError("the order is not submitted before")
        (self.buy_order_book if cancel.order.is_buy else self.sell_order_book).cancel(
            cancel=cancel
        )
        if cancel.placed_at is None:
            raise AssertionError
        self._update_market_price()

        log: CancelLog = CancelLog(
            order_id=cancel.order.order_id,
            market_id=cancel.order.market_id,
            cancel_time=cancel.placed_at,
            order_time=cancel.order.placed_at,
            agent_id=cancel.order.agent_id,
            is_buy=cancel.order.is_buy,
            kind=cancel.order.kind,
            volume=cancel.order.volume,
            price=cancel.order.price,
            ttl=cancel.order.ttl,
        )
        if self.logger is not None:
            log.read_and_write(logger=self.logger)
        return log

    def _update_market_price(self) -> None:
        """update market price. (Internal method)"""
        best_buy_price: Optional[float] = self.get_best_buy_price()
        best_sell_price: Optional[float] = self.get_best_sell_price()
        if best_buy_price is None or best_sell_price is None:
            self._mid_prices[self.time] = None
        else:
            self._mid_prices[self.time] = (
                (best_sell_price + best_buy_price) / 2.0
                if best_sell_price is not None and best_buy_price is not None
                else None
            )
        if self.is_running:
            if self._last_executed_prices[self.time] is not None:
                self._market_prices[self.time] = self._last_executed_prices[self.time]
            elif self._mid_prices[self.time] is not None:
                self._market_prices[self.time] = self._mid_prices[self.time]

    def _execute_orders(
        self, price: float, volume: int, buy_order: Order, sell_order: Order
    ) -> ExecutionLog:
        """execute orders. (Internal method)

        Args:
            price (float): price.
            volume (int): volume.
            buy_order (:class:`pams.order.Order`): buy order.
            sell_order (:class:`pams.order.Order`): sell order.

        Returns:
            :class:`pams.logs.base.CancelLog`: execution log.
        """
        if not self.is_running:
            raise AssertionError("market is not running")
        if buy_order.market_id != self.market_id:
            raise ValueError("buy order is not for this market")
        if sell_order.market_id != self.market_id:
            raise ValueError("sell order is not for this market")

        if buy_order.placed_at is None:
            raise ValueError("buy order is not submitted yet")
        if sell_order.placed_at is None:
            raise ValueError("sell order is not submitted yet")

        if volume <= 0:
            raise AssertionError

        log: ExecutionLog = ExecutionLog(
            market_id=self.market_id,
            time=self.time,
            buy_agent_id=buy_order.agent_id,
            sell_agent_id=sell_order.agent_id,
            buy_order_id=cast(int, buy_order.order_id),
            sell_order_id=cast(int, sell_order.order_id),
            price=price,
            volume=volume,
        )

        self.buy_order_book.change_order_volume(order=buy_order, delta=-volume)
        self.sell_order_book.change_order_volume(order=sell_order, delta=-volume)

        self._last_executed_prices[self.time] = price
        self._executed_volumes[self.time] += volume
        self._executed_total_prices[self.time] += volume * price
        self._update_market_price()

        # ToDo: Agent modification will be handled in simulator
        if self.logger is not None:
            log.read_and_write(logger=self.logger)
        return log

    def _add_order(self, order: Order) -> OrderLog:
        """add order. (Usually, only triggered by runner)

        Args:
            order (:class:`pams.order.Order`): order.

        Returns:
            :class:`pams.logs.base.OrderLog`: order log.
        """
        if order.market_id != self.market_id:
            raise ValueError("order is not for this market")
        if order.placed_at is not None:
            raise ValueError("the order is already submitted")
        if order.order_id is not None:
            raise ValueError("the order is already submitted")
        if order.price is not None and order.price % self.tick_size != 0:
            warnings.warn(
                "order price does not accord to the tick size. price will be modified"
            )
            order.price = (
                self.convert_to_tick_level(price=order.price, is_buy=order.is_buy)
                * self.tick_size
            )
        order.order_id = self._next_order_id
        self._next_order_id += 1
        (self.buy_order_book if order.is_buy else self.sell_order_book).add(order=order)
        if order.placed_at != self.time:
            raise AssertionError
        self._update_market_price()
        if order.is_buy:
            self._n_buy_orders[self.time] += 1
        else:
            self._n_sell_orders[self.time] += 1

        log: OrderLog = OrderLog(
            order_id=order.order_id,
            market_id=order.market_id,
            time=cast(int, order.placed_at),
            agent_id=order.agent_id,
            is_buy=order.is_buy,
            kind=order.kind,
            volume=order.volume,
            price=order.price,
            ttl=order.ttl,
        )
        if self.logger is not None:
            log.read_and_write(logger=self.logger)
        return log

    def remain_executable_orders(self) -> bool:
        """check if there are remain executable orders in this market.

        Returns:
            bool: whether some orders is executable or not.
        """
        if len(self.sell_order_book) == 0:
            return False
        if len(self.buy_order_book) == 0:
            return False
        sell_best: Order = cast(Order, self.sell_order_book.get_best_order())
        buy_best: Order = cast(Order, self.buy_order_book.get_best_order())
        if sell_best.price is not None or buy_best.price is not None:
            if sell_best.price is not None and buy_best.price is not None:
                return sell_best.price <= buy_best.price
            else:
                return True
        else:
            sell_book: Dict[
                Optional[float], int
            ] = self.sell_order_book.get_price_volume()
            buy_book: Dict[
                Optional[float], int
            ] = self.buy_order_book.get_price_volume()
            if None not in sell_book or None not in buy_book:
                raise AssertionError
            if sell_book[None] != buy_book[None]:
                if sell_book[None] < buy_book[None]:
                    additional_required_orders = buy_book[None] - sell_book[None]
                    sell_book.pop(None)
                    return len(sell_book) >= additional_required_orders
                else:
                    additional_required_orders = sell_book[None] - buy_book[None]
                    buy_book.pop(None)
                    return len(buy_book) >= additional_required_orders
            else:
                sell_book.pop(None)
                buy_book.pop(None)
                if len(sell_book) == 0 or len(buy_book) == 0:
                    return False
                return min(list(cast(Dict[float, int], sell_book).keys())) <= max(
                    list(cast(Dict[float, int], buy_book).keys())
                )

    def _execution(self) -> List[ExecutionLog]:
        """execute for market. (Usually, only triggered by runner)

        Returns:
            List[:class:`pams.logs.base.ExecutionLog`]: execution logs.
        """
        if not self.remain_executable_orders():
            return []
        pending: List[Tuple[int, Order, Order]] = []

        popped_buy_orders: List[Order] = []
        popped_sell_orders: List[Order] = []

        buy_order: Order = heapq.heappop(self.buy_order_book.priority_queue)
        popped_buy_orders.append(buy_order)
        sell_order: Order
        buy_order_volume_tmp: int = buy_order.volume
        sell_order_volume_tmp: int = 0
        price: Optional[float] = None
        while True:
            if buy_order_volume_tmp != 0 and sell_order_volume_tmp != 0:
                raise AssertionError
            if buy_order_volume_tmp == 0:
                if len(self.buy_order_book.priority_queue) == 0:
                    break
                buy_order = heapq.heappop(self.buy_order_book.priority_queue)
                popped_buy_orders.append(buy_order)
                buy_order_volume_tmp = buy_order.volume
                if buy_order_volume_tmp == 0:
                    raise AssertionError
            if sell_order_volume_tmp == 0:
                if len(self.sell_order_book.priority_queue) == 0:
                    break
                sell_order = heapq.heappop(self.sell_order_book.priority_queue)
                popped_sell_orders.append(sell_order)
                sell_order_volume_tmp = sell_order.volume
                if sell_order_volume_tmp == 0:
                    raise AssertionError
            if (
                buy_order.price is not None
                and sell_order.price is not None
                and buy_order.price < sell_order.price
            ):
                break
            volume = min(buy_order_volume_tmp, sell_order_volume_tmp)
            if volume == 0:
                raise AssertionError
            buy_order_volume_tmp -= volume
            sell_order_volume_tmp -= volume
            if buy_order_volume_tmp < 0:
                raise AssertionError
            if sell_order_volume_tmp < 0:
                raise AssertionError
            if buy_order.price is None or sell_order.price is None:
                if buy_order.price is None and sell_order.price is None:
                    pending.append((volume, buy_order, sell_order))
                else:
                    price = (
                        buy_order.price
                        if buy_order.price is not None
                        else sell_order.price
                    )
                    pending.append((volume, buy_order, sell_order))
            else:
                if buy_order.placed_at == sell_order.placed_at:
                    if buy_order.order_id is None or sell_order.order_id is None:
                        raise AssertionError
                    if buy_order.order_id < sell_order.order_id:
                        price = buy_order.price
                    elif buy_order.order_id > sell_order.order_id:
                        price = sell_order.price
                    else:
                        raise AssertionError
                else:
                    price = (
                        buy_order.price
                        if cast(int, buy_order.placed_at)
                        < cast(int, sell_order.placed_at)
                        else sell_order.price
                    )
                pending.append((volume, buy_order, sell_order))
        if price is None:
            raise AssertionError
        # TODO: faster impl
        self.buy_order_book.priority_queue = [
            *popped_buy_orders,
            *self.buy_order_book.priority_queue,
        ]
        self.sell_order_book.priority_queue = [
            *popped_sell_orders,
            *self.sell_order_book.priority_queue,
        ]
        heapq.heapify(self.buy_order_book.priority_queue)
        heapq.heapify(self.sell_order_book.priority_queue)
        logs: List[ExecutionLog] = list(
            map(
                lambda x: self._execute_orders(
                    price=cast(float, price),
                    volume=x[0],
                    buy_order=x[1],
                    sell_order=x[2],
                ),
                pending,
            )
        )
        if self.remain_executable_orders():
            raise AssertionError
        if self.logger is not None:
            self.logger.bulk_write(logs=cast(List[Log], logs))
        return logs

    def change_fundamental_price(self, scale: float) -> None:
        """change fundamental price.

        Args:
            scale (float): scale.

        Returns:
            None
        """
        time: int = self.time
        current_fundamental: float = self.get_fundamental_price(time=time)
        new_fundamental: float = current_fundamental * scale
        self._fundamental_prices[time] = new_fundamental
        self.simulator.fundamentals.prices[self.market_id][time] = new_fundamental
        self.simulator.fundamentals._generated_until = time



import random
from typing import Dict
from typing import List
from typing import Optional
from typing import Type
from typing import cast

from .agents.base import Agent
from .agents.high_frequency_agent import HighFrequencyAgent
from .events.base import EventABC
from .events.base import EventHook
from .fundamentals import Fundamentals
from .index_market import IndexMarket
from .logs.base import Logger
from .market import Market
from .session import Session


class Simulator:
    """Simulator class."""

    def __init__(
        self,
        prng: random.Random,
        logger: Optional[Logger] = None,
        fundamental_class: Type[Fundamentals] = Fundamentals,
    ) -> None:
        """initialization.

        Args:
            prng (random.Random): pseudo random number generator for this simulator.
            logger (Logger, Optional): logger for correcting various outputs in one simulation.
                                       logger is usually shared to all classes.
                                       Please note that logger is usually not thread-safe and non-blocking.
            fundamental_class (Type[Fundamentals]): the class that provide fundamental functions for simulator
                                                    (default :class:`pams.fundamentals.Fundamentals`).

        Note:
             `prng` should not be shared with other classes and be used only in this class.
             It is because sometimes agent process runs one of parallelized threads.
        """
        self._prng = prng
        self.logger: Optional[Logger] = logger
        if self.logger is not None:
            self.logger._set_simulator(simulator=self)

        self.n_events: int = 0
        self.events: List[EventABC] = []
        self.id2event: Dict[int, EventABC] = {}
        self.event_hooks: List[EventHook] = []
        self.events_dict: Dict[str, Dict[Optional[int], List[EventHook]]] = {
            "order_before": {},
            "order_after": {},
            "cancel_before": {},
            "cancel_after": {},
            "execution_after": {},
            "session_before": {},
            "session_after": {},
            "market_before": {},
            "market_after": {},
        }
        self.name2event: Dict[str, EventABC] = {}

        self.n_agents: int = 0
        self.agents: List[Agent] = []
        self.high_frequency_agents: List[Agent] = []
        self.normal_frequency_agents: List[Agent] = []
        self.id2agent: Dict[int, Agent] = {}
        self.name2agent: Dict[str, Agent] = {}
        self.agents_group_name2agent: Dict[str, List[Agent]] = {}

        self.n_markets: int = 0
        self.markets: List[Market] = []
        self.id2market: Dict[int, Market] = {}
        self.name2market: Dict[str, Market] = {}
        self.markets_group_name2market: Dict[str, List[Market]] = {}

        self.fundamentals = fundamental_class(
            prng=random.Random(self._prng.randint(0, 2**31))
        )

        self.n_sessions: int = 0
        self.sessions: List[Session] = []
        self.id2session: Dict[int, Session] = {}
        self.name2session: Dict[str, Session] = {}
        self.current_session: Optional[Session] = None

    def _add_event(self, event_hook: EventHook) -> None:
        """add event to the simulator. (Usually, this is called from runner.)

        Args:
            event_hook (:class:`pams.events.base.EventHook`): event hook.

        Returns:
            None
        """
        if event_hook in self.event_hooks:
            raise ValueError("event_hook is already registered")
        event = event_hook.event
        if event_hook.event not in self.events:
            self.events.append(event)
        self.n_events += 1
        if event_hook.event.event_id not in self.id2event:
            self.id2event[event.event_id] = event
        if event_hook.event.name not in self.name2event:
            self.name2event[event.name] = event
        self.event_hooks.append(event_hook)
        register_name: str = event_hook.hook_type + (
            "_before" if event_hook.is_before else "_after"
        )
        times: List[Optional[int]] = (
            cast(List[Optional[int]], event_hook.time)
            if event_hook.time is not None
            else cast(List[Optional[int]], [None])
        )
        for time_ in times:
            if time_ not in self.events_dict[register_name]:
                self.events_dict[register_name][time_] = []
            self.events_dict[register_name][time_].append(event_hook)

    def _add_market(self, market: Market, group_name: Optional[str] = None) -> None:
        """add market to the simulator. (Usually, this is called from runner.)

        Args:
            market (:class:`pamd.market.Market`): market.
            group_name (str, Optional): group name for market (default None).

        Returns:
            None
        """
        if market in self.markets:
            raise ValueError("market is already registered")
        if market.market_id in self.id2market:
            raise ValueError(f"market_id {market.market_id} is duplicated")
        if market.name in self.name2market:
            raise ValueError(f"market name {market.name} is duplicate")
        self.markets.append(market)
        self.n_markets += 1
        self.id2market[market.market_id] = market
        self.name2market[market.name] = market
        if group_name is not None:
            if group_name not in self.markets_group_name2market:
                self.markets_group_name2market[group_name] = []
            self.markets_group_name2market[group_name].append(market)

    def _add_agent(self, agent: Agent, group_name: Optional[str] = None) -> None:
        """add agent to the simulator. (Usually, this is called from runner.)

        Args:
            agent (:class:`pams.agents.base.Agent`): agent.
            group_name (str, Optional): group name for agent (default None).

        Returns:
            None
        """
        if agent in self.agents:
            raise ValueError("agent is already registered")
        if agent.agent_id in self.id2agent:
            raise ValueError(f"agent_id {agent.agent_id} is duplicated")
        if agent.name in self.name2agent:
            raise ValueError(f"agent name {agent.name} is duplicate")
        self.agents.append(agent)
        self.n_agents += 1
        self.id2agent[agent.agent_id] = agent
        self.name2agent[agent.name] = agent
        if isinstance(agent, HighFrequencyAgent):
            self.high_frequency_agents.append(agent)
        else:
            self.normal_frequency_agents.append(agent)
        if group_name is not None:
            if group_name not in self.agents_group_name2agent:
                self.agents_group_name2agent[group_name] = []
            self.agents_group_name2agent[group_name].append(agent)

    def _add_session(self, session: Session) -> None:
        """add session to the simulator. (Usually, this is called from runner.)

        Args:
            session (:class:`pams.session.Session`): session.

        Returns:
            None
        """
        if session in self.sessions:
            raise ValueError("session is already registered")
        if session.session_id in self.id2session:
            raise ValueError(f"session_id {session.session_id} is duplicated")
        if session.name in self.name2session:
            raise ValueError(f"session name {session.name} is duplicate")
        self.sessions.append(session)
        self.n_sessions += 1
        self.id2session[session.session_id] = session
        self.name2session[session.name] = session

    def _update_time_on_market(self, market: Market) -> None:
        """update time on the market. (Usually, this is called from runner.)

        Args:
            market (:class:`pams.market.Market`): market.

        Returns:
            None

        Notes:
            be careful index matket have to be update after component markets.
            Technically, the fundamental values for components markets can be calculated beforehand, but not allowed to avoid future data leakage.
        """
        if not isinstance(market, IndexMarket):
            market._update_time(
                next_fundamental_price=self.fundamentals.get_fundamental_price(
                    market_id=market.market_id, time=market.get_time() + 1
                )
            )
        else:
            market._update_time(
                next_fundamental_price=market.compute_fundamental_index(
                    time=market.get_time() + 1
                )
            )

    def _update_times_on_markets(self, markets: List[Market]) -> None:
        """update times on markets. (Usually, this is called from runner.)

        Args:
            markets (List[:class:`pams.market.Market`]): list of markets.

        Returns:
            None
        """
        for market in filter(lambda x: not isinstance(x, IndexMarket), markets):
            self._update_time_on_market(market=market)
        for market in filter(lambda x: isinstance(x, IndexMarket), markets):
            self._update_time_on_market(market=market)

    def _update_agents_for_execution(
        self, execution_logs: List["ExecutionLog"]  # type: ignore  # NOQA
    ) -> None:
        """update agents for execution. (Usually, this is called from runner.)

        Args:
            execution_logs (List["ExecutionLog"]): execution logs.

        Returns:
            None
        """
        for log in execution_logs:
            buy_agent: Agent = self.id2agent[log.buy_agent_id]
            sell_agent: Agent = self.id2agent[log.sell_agent_id]
            price: float = log.price
            volume: int = log.volume
            market_id: int = log.market_id
            buy_agent.cash_amount -= price * volume
            sell_agent.cash_amount += price * volume
            buy_agent.asset_volumes[market_id] += volume
            sell_agent.asset_volumes[market_id] -= volume

    def _check_event_class_and_instance(
        self,
        check_object: object,
        class_requirement: Optional[Type] = None,
        instance_requirement: Optional[object] = None,
    ) -> bool:
        """check event class and instance. (Usually, this is called from runner.)

        Args:
            check_object (object): object for check.
            class_requirement (Type, Optional): class requirement.
            instance_requirement (object, Optional): instance requirement.

        Returns:
            bool: whether the event class or instance meet the requirements.
        """
        if class_requirement is not None:
            if not isinstance(check_object, class_requirement):
                return False
        if instance_requirement is not None:
            if instance_requirement != check_object:
                return False
        return True

    def _trigger_event_before_order(self, order: "Order") -> None:  # type: ignore  # NOQA
        """trigger event before order. (Usually, this is called from runner.)

        Args:
            order (Order): the order before the event.

        Returns:
            None
        """
        time: int = order.placed_at
        event_hooks = self.events_dict["order_before"]
        target_event_hooks: List[EventHook] = []
        if None in event_hooks:
            target_event_hooks.extend(event_hooks[None])
        if time in event_hooks:
            target_event_hooks.extend(event_hooks[time])
        for event_hook in target_event_hooks:
            event_hook.event.hooked_before_order(simulator=self, order=order)

    def _trigger_event_after_order(self, order_log: "OrderLog") -> None:  # type: ignore  # NOQA
        """trigger event after order. (Usually, this is called from runner.)

        Args:
            order_log (OrderLog): the order log after the event.

        Returns:
            None
        """
        time: int = order_log.time
        event_hooks = self.events_dict["order_after"]
        target_event_hooks: List[EventHook] = []
        if None in event_hooks:
            target_event_hooks.extend(event_hooks[None])
        if time in event_hooks:
            target_event_hooks.extend(event_hooks[time])
        for event_hook in target_event_hooks:
            event_hook.event.hooked_after_order(simulator=self, order_log=order_log)

    def _trigger_event_before_cancel(self, cancel: "Cancel") -> None:  # type: ignore  # NOQA
        """trigger event before cancel. (Usually, this is called from runner.)

        Args:
            cancel (Cancel): the cancel order before the event.

        Returns:
            None
        """
        time: int = cancel.placed_at
        event_hooks = self.events_dict["cancel_before"]
        target_event_hooks: List[EventHook] = []
        if None in event_hooks:
            target_event_hooks.extend(event_hooks[None])
        if time in event_hooks:
            target_event_hooks.extend(event_hooks[time])
        for event_hook in target_event_hooks:
            event_hook.event.hooked_before_cancel(simulator=self, cancel=cancel)

    def _trigger_event_after_cancel(self, cancel_log: "CancelLog") -> None:  # type: ignore  # NOQA
        """trigger event after cancel. (Usually, this is called from runner.)

        Args:
            cancel_log (CancelLog): the cancel order log after the event.

        Returns:
            None
        """
        time: int = cancel_log.cancel_time
        event_hooks = self.events_dict["cancel_after"]
        target_event_hooks: List[EventHook] = []
        if None in event_hooks:
            target_event_hooks.extend(event_hooks[None])
        if time in event_hooks:
            target_event_hooks.extend(event_hooks[time])
        for event_hook in target_event_hooks:
            event_hook.event.hooked_after_cancel(simulator=self, cancel_log=cancel_log)

    def _trigger_event_after_execution(self, execution_log: "ExecutionLog") -> None:  # type: ignore  # NOQA
        """trigger event after execution. (Usually, this is called from runner.)

        Args:
            execution_log (ExecutionLog): the execution log after the event.

        Returns:
            None
        """
        time: int = execution_log.time
        event_hooks = self.events_dict["execution_after"]
        target_event_hooks: List[EventHook] = []
        if None in event_hooks:
            target_event_hooks.extend(event_hooks[None])
        if time in event_hooks:
            target_event_hooks.extend(event_hooks[time])
        for event_hook in target_event_hooks:
            event_hook.event.hooked_after_execution(
                simulator=self, execution_log=execution_log
            )

    def _trigger_event_before_session(self, session: "Session") -> None:  # type: ignore
        """trigger event before session. (Usually, this is called from runner.)

        Args:
            session (Session): the session before the event.

        Returns:
            None
        """
        time: int = session.session_start_time
        event_hooks = self.events_dict["session_before"]
        target_event_hooks: List[EventHook] = []
        if None in event_hooks:
            target_event_hooks.extend(event_hooks[None])
        if time in event_hooks:
            target_event_hooks.extend(event_hooks[time])
        for event_hook in target_event_hooks:
            event_hook.event.hooked_before_session(simulator=self, session=session)

    def _trigger_event_after_session(self, session: "Session") -> None:  # type: ignore
        """trigger event after session. (Usually, this is called from runner.)

        Args:
            session (Session): the session after the event.

        Returns:
            None
        """
        time: int = session.session_start_time + session.iteration_steps - 1
        event_hooks = self.events_dict["session_after"]
        target_event_hooks: List[EventHook] = []
        if None in event_hooks:
            target_event_hooks.extend(event_hooks[None])
        if time in event_hooks:
            target_event_hooks.extend(event_hooks[time])
        for event_hook in target_event_hooks:
            event_hook.event.hooked_after_session(simulator=self, session=session)

    def _trigger_event_before_step_for_market(self, market: "Market") -> None:  # type: ignore
        """trigger event before step for market. (Usually, this is called from runner.)

        Args:
            market (Market): the market before the event.

        Returns:
            None
        """
        time: int = market.get_time()
        event_hooks = self.events_dict["market_before"]
        target_event_hooks: List[EventHook] = []
        if None in event_hooks:
            target_event_hooks.extend(event_hooks[None])
        if time in event_hooks:
            target_event_hooks.extend(event_hooks[time])
        for event_hook in target_event_hooks:
            if self._check_event_class_and_instance(
                check_object=market,
                class_requirement=event_hook.specific_class,
                instance_requirement=event_hook.specific_instance,
            ):
                event_hook.event.hooked_before_step_for_market(
                    simulator=self, market=market
                )

    def _trigger_event_after_step_for_market(self, market: "Market") -> None:  # type: ignore
        """trigger event after step for market. (Usually, this is called from runner.)

        Args:
            market (Market): the market after the event.

        Returns:
            None
        """
        time: int = market.get_time()
        event_hooks = self.events_dict["market_after"]
        target_event_hooks: List[EventHook] = []
        if None in event_hooks:
            target_event_hooks.extend(event_hooks[None])
        if time in event_hooks:
            target_event_hooks.extend(event_hooks[time])
        for event_hook in target_event_hooks:
            if self._check_event_class_and_instance(
                check_object=market,
                class_requirement=event_hook.specific_class,
                instance_requirement=event_hook.specific_instance,
            ):
                event_hook.event.hooked_after_step_for_market(
                    simulator=self, market=market
                )

    # ToDo get_xxx_by_name



import random
import warnings
from typing import Any
from typing import Dict
from typing import List
from typing import Optional

from ..market import Market
from ..order import Order
from ..session import Session
from ..simulator import Simulator
from .base import EventABC
from .base import EventHook


class PriceLimitRule(EventABC):
    """This limits the price range.

    The order having the price that is out of the price range, the price is overridden to the edge of range.
    This event is only called via :func:`hooked_before_order` at designated step.
    """

    def __init__(
        self,
        event_id: int,
        prng: random.Random,
        session: Session,
        simulator: Simulator,
        name: str,
    ) -> None:
        super().__init__(
            event_id=event_id,
            prng=prng,
            session=session,
            simulator=simulator,
            name=name,
        )
        self.target_markets: Dict[str, Market] = {}
        self.is_enabled: bool = True
        self.activation_count: int = 0
        self.trigger_change_rate: float = 0.0

    def setup(self, settings: Dict[str, Any], *args, **kwargs) -> None:  # type: ignore  # NOQA
        """event setup. Usually be called from simulator/runner automatically.

        Args:
            settings (Dict[str, Any]): agent configuration. Usually, automatically set from json config of simulator.
                                       This must include the parameters "targetMarkets" and "triggerChangeRate".
                                       This can include the parameters "enabled".
                                       The parameter "referenceMarket" is obsoleted.

        Returns:
            None
        """
        if "referenceMarket" in settings:
            warnings.warn("referenceMarket is obsoleted")
        if "targetMarkets" not in settings:
            raise ValueError("targetMarkets is required for PriceLimitRule")
        if not isinstance(settings["targetMarkets"], list):
            raise ValueError("targetMarkets must be list")
        for market_name in settings["targetMarkets"]:
            if not isinstance(market_name, str):
                raise ValueError("constituent of targetMarkets have to be string")
            if market_name not in self.simulator.name2market:
                raise ValueError(f"{market_name} does not exist")
            market: Market = self.simulator.name2market[market_name]
            self.target_markets[market_name] = market
        if "triggerChangeRate" not in settings:
            raise ValueError("triggerChangeRate is required for PriceLimitRule")
        if not isinstance(settings["triggerChangeRate"], float):
            raise ValueError("triggerChangeRate have to be float.")
        self.trigger_change_rate = settings["triggerChangeRate"]
        if "enabled" in settings:
            self.is_enabled = settings["enabled"]

    def hook_registration(self) -> List[EventHook]:
        if self.is_enabled:
            event_hook = EventHook(
                event=self, hook_type="order", is_before=True, time=None
            )
            return [event_hook]
        else:
            return []

    def get_limited_price(self, order: Order, market: Market) -> Optional[float]:
        """Calculate the limited price for an order.

        Args:
            order (Order): order whose price is calculated
            market (Market): market that order belongs to

        Returns:
            Optional[float]: price after price limit. If the input order is market order, the return become None (market order).
        """
        reference_price = market.get_market_price(0)
        if market not in self.target_markets.values():
            raise AssertionError
        if order.price is None:
            return order.price
        order_price: float = order.price
        price_change: float = order_price - reference_price
        threshold_change: float = reference_price * self.trigger_change_rate
        if abs(price_change) >= abs(threshold_change):
            max_price: float = reference_price * (1 + self.trigger_change_rate)
            min_price: float = reference_price * (1 - self.trigger_change_rate)
            limited_price: float = min(max(order_price, min_price), max_price)
            return limited_price
        return order_price

    def hooked_before_order(self, simulator: Simulator, order: Order) -> None:
        new_price: Optional[float] = self.get_limited_price(
            order, simulator.id2market[order.market_id]
        )
        if order.price != new_price:
            self.activation_count += 1
        order.price = new_price


PriceLimitRule.hook_registration.__doc__ = EventABC.hook_registration.__doc__
PriceLimitRule.hooked_before_order.__doc__ = EventABC.hooked_before_order.__doc__



import random
import warnings
from typing import Any
from typing import Dict
from typing import List
from typing import Optional
from typing import cast

from .logs import Logger
from .market import Market


class IndexMarket(Market):
    """Index of market.

    This class inherits from the :class:`pams.market.Market` class.
    """

    def __init__(
        self,
        market_id: int,
        prng: random.Random,
        simulator: "Simulator",  # type: ignore  # NOQA
        name: str,
        logger: Optional[Logger] = None,
    ):
        super().__init__(
            market_id=market_id,
            prng=prng,
            simulator=simulator,
            name=name,
            logger=logger,
        )
        self._components: List[Market] = []

    def setup(self, settings: Dict[str, Any], *args, **kwargs) -> None:  # type: ignore
        """setup market configuration from setting format.

        Args:
            settings (Dict[str, Any]): market configuration. Usually, automatically set from json config of simulator.
                                       This must include the parameter "markets".
                                       This should include the parameter "requires".

        Returns:
            None
        """
        super(IndexMarket, self).setup(settings, *args, **kwargs)
        if "markets" not in settings:
            raise ValueError("markets is required for index markets as components")
        if "requires" in settings:
            warnings.warn("requires in index market settings is no longer required")
        for market_name in settings["markets"]:
            market: Market = self.simulator.name2market[market_name]
            self._add_market(market=market)

    def _add_market(self, market: Market) -> None:
        """add market. (Internal method)

        Args:
            market (:class:`pams.market.Market`): market.

        Returns:
            None
        """
        if market in self._components:
            raise ValueError("market is already registered as components")
        if market.outstanding_shares is None:
            raise AssertionError(
                "outstandingShares is required in component market setting"
            )
        self._components.append(market)

    def _add_markets(self, markets: List[Market]) -> None:
        """add markets. (Internal method)

        Args:
            markets (List[:class:`pams.market.Market`]): list of market.

        Returns:
            None
        """
        for market in markets:
            self._add_market(market=market)

    def compute_fundamental_index(self, time: Optional[int] = None) -> float:
        """compute fundamental index.

        Args:
            time (int, Optional): time step.

        Returns:
            float: fundamental index.

        .. note::
            In an index market, there are two types of fundamental values:
             1. fundamental value set from outside such as runner (:func:`pams.index_market.IndexMarket.compute_fundamental_index`)
             2. fundamental value calculated from components' fundamental value
                (:func:`pams.index_market.IndexMarket.get_fundamental_index`,
                 :func:`pams.index_market.IndexMarket.get_fundamental_price`)
            In usual usage, those become the same. But, in some special usage, it could be differ.
            This method return 1st one.
        """
        if time is None:
            time = self.get_time()
        total_value: float = 0
        total_shares: int = 0
        for market in self._components:
            outstanding_shares = cast(int, market.outstanding_shares)
            total_value += market.get_fundamental_price(time=time) * outstanding_shares
            total_shares += cast(int, outstanding_shares)
        return total_value / total_shares

    def compute_market_index(self, time: Optional[int] = None) -> float:
        """compute market index.

        Args:
            time (int, Optional): time step.

        Returns:
            float: market index.
        """
        if time is None:
            time = self.get_time()
        total_value: float = 0
        total_shares: int = 0
        for market in self._components:
            outstanding_shares = cast(int, market.outstanding_shares)
            total_value += market.get_market_price(time=time) * outstanding_shares
            total_shares += outstanding_shares
        return total_value / total_shares

    def get_components(self) -> List[Market]:
        """get components.

        Returns:
            List[:class:`pams.market.Market`]: list of components.
        """
        return self._components

    def is_all_markets_running(self) -> bool:
        """get whether all markets is running or not.

        Returns:
            bool: whether all markets is running or not.
        """
        return sum(map(lambda x: not x.is_running, self._components)) == 0

    def get_fundamental_index(self, time: Optional[int] = None) -> float:
        """get fundamental index.

        Args:
            time (int, Optional): time step.

        Returns:
            float: fundamental index.

        .. note::
            In an index market, there are two types of fundamental values:
             1. fundamental value set from outside such as runner (:func:`pams.index_market.IndexMarket.compute_fundamental_index`)
             2. fundamental value calculated from components' fundamental value
                (:func:`pams.index_market.IndexMarket.get_fundamental_index`,
                 :func:`pams.index_market.IndexMarket.get_fundamental_price`)
            In usual usage, those become the same. But, in some special usage, it could be differ.
            This method return 2nd one.
        """
        return self.get_fundamental_price(time=time)

    def get_market_index(self, time: Optional[int] = None) -> float:
        """get computed market index.

        Args:
            time (int, Optional): time step.

        Returns:
            float: market index.
        """
        return self.compute_market_index(time=time)

    def get_index(self, time: Optional[int] = None) -> float:
        """get market index.

        Args:
            time (int, Optional): time step.

        Returns:
            float: market index.
        """
        return self.get_market_index(time=time)



from typing import List
from typing import Optional
from typing import Type


def find_class(name: str, optional_class_list: Optional[List[Type]] = None) -> Type:
    """find class from pams name spaces.

    .. seealso::
        If you want to use user-defined classes, please use :class:`pams.runners.class_register`.

    Args:
        name (str): class name.
        optional_class_list (List[Type], Optional): optional class list.

    Returns:
        Type: class type.
    """
    _1 = __import__("pams", globals(), locals())
    _2 = __import__("pams.agents", globals(), locals(), ["*"])
    _3 = __import__("pams.events", globals(), locals(), ["*"])
    _4 = __import__("pams.logs", globals(), locals(), ["*"])
    _5 = __import__("pams.utils", globals(), locals(), ["*"])

    candidates_spaces = [*globals().values(), *locals().values()]

    object_class_candidates = [
        getattr(m, name) for m in candidates_spaces if hasattr(m, name)
    ]
    if optional_class_list is not None:
        object_class_candidates.extend(
            [x for x in optional_class_list if x.__name__ == name]
        )
    if len(object_class_candidates) != 1:
        raise AttributeError(
            f"class for {name} is found {len(object_class_candidates)} times"
        )
    object_class: Type = object_class_candidates[0]
    return object_class



import heapq
from typing import Dict
from typing import List
from typing import Optional

from .order import Cancel
from .order import Order


class OrderBook:
    """Order book class."""

    def __init__(self, is_buy: bool) -> None:
        """initialization.

        Args:
            is_buy (bool): whether it is a buy order or not.

        Returns:
            None
        """
        self.priority_queue: List[Order] = []
        heapq.heapify(self.priority_queue)
        self.time: int = 0
        self.is_buy = is_buy
        self.expire_time_list: Dict[int, List[Order]] = {}

    def __repr__(self) -> str:
        return f"<{self.__class__.__module__}.{self.__class__.__name__} | is_buy={self.is_buy}>"

    def add(self, order: Order) -> None:
        """add the book of order.

        Args:
            order (:class:`pams.order.Order`): order.

        Returns:
            None
        """
        if order.is_buy != self.is_buy:
            raise ValueError("buy/sell is incorrect")
        order.placed_at = self.time
        heapq.heappush(self.priority_queue, order)
        if order.ttl is not None:
            expiration_time = order.placed_at + order.ttl
            if expiration_time not in self.expire_time_list:
                self.expire_time_list[expiration_time] = []
            self.expire_time_list[expiration_time].append(order)

    def _remove(self, order: Order) -> None:
        """remove the book of order. (Internal method. Usually, it is not called from the outside of this class.)

        Args:
            order (:class:`pams.order.Order`): order.

        Returns:
            None
        """
        if order == self.priority_queue[0]:
            x = heapq.heappop(self.priority_queue)
            assert x == order
        else:
            self.priority_queue.remove(order)
            heapq.heapify(self.priority_queue)
        if order.placed_at is None:
            raise AssertionError("the order is not yet placed")
        if order.ttl is not None:
            expiration_time = order.placed_at + order.ttl
            self.expire_time_list[expiration_time].remove(order)

    def cancel(self, cancel: Cancel) -> None:
        """cancel the book of order.

        Args:
            cancel (:class:`pams.order.Cancel`): cancel order.

        Returns:
            None
        """
        cancel.order.is_canceled = True
        cancel.placed_at = self.time
        if cancel.order in self.priority_queue:
            # in case that order is executed before canceling.
            self._remove(cancel.order)

    def get_best_order(self) -> Optional[Order]:
        """get the order with the highest priority.

        Returns:
            :class:`pams.order.Order`, Optional: the order with the highest priority.
        """
        if len(self.priority_queue) > 0:
            return self.priority_queue[0]
        else:
            return None

    def get_best_price(self) -> Optional[float]:
        """get the order price with the highest priority.

        Returns:
            float, Optional: the order price with the highest priority.
        """
        if len(self.priority_queue) > 0:
            return self.priority_queue[0].price
        else:
            return None

    def change_order_volume(self, order: Order, delta: int) -> None:
        """change order volume.

        Args:
            order (:class:`pams.order.Order`): order.
            delta (int): amount of volume change.

        Returns:
            None
        """
        order.volume += delta
        # ToDo: check if volume is non-negative
        if order.volume == 0:
            self._remove(order=order)
        if order.volume < 0:
            raise AssertionError

    def _check_expired_orders(self) -> None:
        """check and delete expired orders. (Internal Method)"""
        delete_orders: List[Order] = sum(
            [value for key, value in self.expire_time_list.items() if key < self.time],
            [],
        )
        delete_keys: List[int] = [
            key for key, value in self.expire_time_list.items() if key < self.time
        ]
        if len(delete_orders) == 0:
            return
        # TODO: think better sorting in the following 3 lines
        for delete_order in delete_orders:
            self.priority_queue.remove(delete_order)
        heapq.heapify(self.priority_queue)
        for key in delete_keys:
            self.expire_time_list.pop(key)

    def _set_time(self, time: int) -> None:
        """set time step. (Usually, it is called from market.)

        Args:
            time (int): time step.

        Returns:
            None
        """
        self.time = time
        self._check_expired_orders()

    def _update_time(self) -> None:
        """update time. (Usually, it is called from market.)
        Advance the time step and check expired orders.
        """
        self.time += 1
        self._check_expired_orders()

    def __len__(self) -> int:
        """get length of the order queue.

        Returns:
            int: length of the order queue.
        """
        return len(self.priority_queue)

    def get_price_volume(self) -> Dict[Optional[float], int]:
        """get price and volume (order book).

        Returns:
            Dict[Optional[float], int]: order book dict. Dict key is order price and the value is volumes.
        """
        keys: List[Optional[float]] = list(
            set(map(lambda x: x.price, self.priority_queue))
        )
        has_market_order: bool = None in keys
        if has_market_order:
            keys.remove(None)
        keys.sort(reverse=self.is_buy)
        if has_market_order:
            keys.insert(0, None)
        result: Dict[Optional[float], int] = dict(
            [
                (
                    key,
                    sum(
                        [
                            order.volume
                            for order in self.priority_queue
                            if order.price == key
                        ]
                    ),
                )
                for key in keys
            ]
        )
        return result



from typing import Dict
from typing import List
from typing import Optional


def json_extends(
    whole_json: Dict,
    parent_name: str,
    target_json: Dict,
    excludes_fields: Optional[List[str]] = None,
) -> Dict:
    """extend target json. If "extends" keys are included in target_json, this try to extend the json dict recursively.
     For the extension, the value for "extends" field  is found in whole_json and target_json is extended using the dict under
     the found part of whole_json. This process is recursively repeated. "parent_name" is only used for avoiding circle
     extension of target_json and it should be parent key of taregt_json.

    Args:
        whole_json (Dict): whole of json.
        parent_name (str): parent name of target_json.
        target_json (Dict): target json.
        excludes_fields (List[str], Optional): exclude list from json field.

    Returns:
        Dict: json output.

    Examples:
        >>> from pams.utils.json_extends import json_extends
        >>> whole_json = {"parent": {"parent_attr": "1", "override_attr": "1"}}
        >>> target_json = {"extends": "parent", "child_attr": "0", "override_attr": "0"}
        >>> json_extends(whole_json=whole_json, parent_name="child", target_json=target_json)
        {'parent_attr': '1', 'override_attr': '0', 'child_attr': '0'}
    """
    # ToDo: add a code example to doc.
    excludes_fields_: List[str] = excludes_fields if excludes_fields is not None else []
    results = target_json.copy()
    extending_history: List[str] = [parent_name]
    while "extends" in results:
        extends_class: str = results["extends"]
        results.pop("extends")
        if extends_class not in whole_json:
            raise ValueError(
                f"{parent_name} cannot extends {extends_class} because {extends_class} is missing"
            )
        if extends_class in extending_history:
            raise ValueError(f"{parent_name} has extending loop")
        extending_history.append(extends_class)
        extending_dict: Dict = whole_json[extends_class]
        results = dict(
            [
                (key, value)
                for key, value in extending_dict.items()
                if key not in excludes_fields_
            ],
            **results,
        )
    return results



import random
import warnings
from typing import Any
from typing import Dict
from typing import List

from ..logs import ExecutionLog
from ..market import Market
from ..session import Session
from ..simulator import Simulator
from .base import EventABC
from .base import EventHook


class TradingHaltRule(EventABC):
    """A trading halt is a market regulation that suspends the trading of some assets.

    When the one of targetMarkets violate halting line, all of them will be halted.
    If you want to halt them separately, please make event separately.
    The current implementation sets :func:`pams.market.Market.is_running` = false when the price changed beyond the prespecified threshold range.
    """

    def __init__(
        self,
        event_id: int,
        prng: random.Random,
        session: Session,
        simulator: Simulator,
        name: str,
    ) -> None:
        super().__init__(
            event_id=event_id,
            prng=prng,
            session=session,
            simulator=simulator,
            name=name,
        )
        self.is_enabled: bool = True
        self.halting_time_length: int = 1
        self.halting_time_started: int = 0
        self.activation_count: int = 0
        self.target_markets: Dict[str, Market] = {}
        self.trigger_change_rate: float = 0.0

    def setup(self, settings: Dict[str, Any], *args, **kwargs) -> None:  # type: ignore  # NOQA
        """event setup. Usually be called from simulator/runner automatically.

        Args:
            settings (Dict[str, Any]): agent configuration. Usually, automatically set from json config of simulator.
                                       This must include the parameters "targetMarkets" and "triggerChangeRate".
                                       This can include the parameters "haltingTimeLength" and "enabled".
                                       The parameter "referenceMarket" is obsoleted.

        Returns:
            None
        """
        if "referenceMarket" in settings:
            warnings.warn("referenceMarket is obsolete")
        if "targetMarkets" not in settings:
            raise ValueError("targetMarkets is required for TradingHaltRule")
        if not isinstance(settings["targetMarkets"], list):
            raise ValueError("targetMarkets must be list")
        for market_name in settings["targetMarkets"]:
            if not isinstance(market_name, str):
                raise ValueError("constituent of targetMarkets must be string")
            if market_name not in self.simulator.name2market:
                raise ValueError(f"{market_name} does not exist")
            market: Market = self.simulator.name2market[market_name]
            self.target_markets[market_name] = market
        if "triggerChangeRate" not in settings:
            raise ValueError("triggerChangeRate is required for TradingHaltRule")
        if not isinstance(settings["triggerChangeRate"], float):
            raise ValueError("triggerChangeRate have to be float")
        self.trigger_change_rate = settings["triggerChangeRate"]
        if "haltingTimeLength" not in settings:
            raise ValueError("haltingTimeLength is required")
        if not isinstance(settings["haltingTimeLength"], int):
            raise ValueError("haltingTimeLength have to be int")
        self.halting_time_length = settings["haltingTimeLength"]
        if "enabled" in settings:
            self.is_enabled = settings["enabled"]

    def hook_registration(self) -> List[EventHook]:
        if self.is_enabled:
            event_hooked_after_execution = EventHook(
                event=self, hook_type="execution", is_before=False
            )
            event_hooked_before_step_for_market = []
            for market in self.target_markets.values():
                event_hooked_before_step_for_market.append(
                    EventHook(
                        event=self,
                        hook_type="market",
                        is_before=True,
                        specific_instance=market,
                    )
                )
            return [event_hooked_after_execution] + event_hooked_before_step_for_market
        else:
            return []

    def hooked_after_execution(
        self, simulator: Simulator, execution_log: ExecutionLog
    ) -> None:
        """event to stop the trading."""
        market: Market = simulator.id2market[execution_log.market_id]
        reference_price = market.get_market_price(0)
        if market.is_running:
            price_change: float = reference_price - market.get_market_price()
            threshold_change: float = (
                reference_price * self.trigger_change_rate * (self.activation_count + 1)
            )
            if abs(price_change) >= abs(threshold_change):
                for m in self.target_markets.values():
                    if m == market:
                        m._is_running = False
                        self.halting_time_started = m.time
                        self.activation_count += 1
                        if simulator.current_session is None:
                            raise AssertionError
                        simulator.current_session.with_order_execution = False

    def hooked_before_step_for_market(
        self, simulator: Simulator, market: Market
    ) -> None:
        """event to start the trading."""
        # TODO: when halting is continued over session
        if market.get_time() > self.halting_time_started + self.halting_time_length:
            for m in self.target_markets.values():
                if m == market:
                    if simulator.current_session is None:
                        raise AssertionError
                    simulator.current_session.with_order_execution = True
                    m._is_running = True
                    self.halting_time_started = 0


TradingHaltRule.hook_registration.__doc__ = EventABC.hook_registration.__doc__
TradingHaltRule.hooked_after_execution.__doc__ = EventABC.hooked_after_execution.__doc__
TradingHaltRule.hooked_before_step_for_market.__doc__ = (
    EventABC.hooked_before_step_for_market.__doc__
)
